import typing as tp

if tp.TYPE_CHECKING:
    from meator.entities import Request
    from meator.interfaces.handlers import IHandler
    from meator.interfaces.middleware import IMiddleware

T = tp.TypeVar("T")

CallNextType: tp.TypeAlias = "IHandler[Request, T] | IMiddleware"



import typing as tp

from meator.entities import Query

from .request import IHandler

HRes = tp.TypeVar("HRes")
QType = tp.TypeVar("QType", bound=Query)


class IQueryHandler(IHandler[QType, HRes]): ...



import typing as tp

from meator.entities import Command

from .request import IHandler

HRes = tp.TypeVar("HRes")
CType = tp.TypeVar("CType", bound=Command)


class ICommandHandler(IHandler[CType, HRes]): ...



import abc
import typing as tp

from meator.entities import Request
from meator.interfaces.handlers.request import IHandler

Res = tp.TypeVar("Res")
Req = tp.TypeVar("Req", bound=Request)


class IObserver(abc.ABC):
    @abc.abstractmethod
    def register(self, request: type[Req], handlers: tp.Sequence[IHandler[Req, Res]]) -> None:
        """Register handler for request."""

    @abc.abstractmethod
    async def handle(self, request: Request[None]) -> None:
        """Handle request."""



from .command import ICommandHandler
from .event import IEventHandler
from .query import IQueryHandler
from .request import IHandler

__all__ = ("ICommandHandler", "IQueryHandler", "IEventHandler", "IHandler")



import typing as tp

from meator._constants import _sentinel

T = tp.TypeVar("T")


def get_default_sentinel(value: T, default_value: T) -> T:
    return value if value is not _sentinel else default_value



import typing as tp
from typing import TypeVar

from meator._constants import _sentinel
from meator._utils.sentinel_default import get_default_sentinel
from meator.dispatchers.errors import HandlerNotFoundError
from meator.entities.request import Request
from meator.interfaces.dispatcher import IDispatcher
from meator.interfaces.handlers.request import IHandler
from meator.interfaces.middleware import IMiddleware
from meator.middlewares import wrap_handler_with_middleware

Res = TypeVar("Res")
RType = TypeVar("RType", bound=Request)


class BaseDispatcherImpl(IDispatcher):
    def __init__(
        self,
        *,
        middlewares: tp.Sequence[IMiddleware] = _sentinel,
        handlers: dict[type[Request], IHandler] = _sentinel,
    ) -> None:
        self._middlewares = get_default_sentinel(middlewares, [])
        self._handlers = get_default_sentinel(handlers, {})

    def register(self, request: type[RType], handler: IHandler[RType, Res]) -> None:
        self._handlers[request] = handler

    async def handle(self, request: Request[Res]) -> Res:
        handler = self._get_handler(type(request))
        return await wrap_handler_with_middleware(self._middlewares, handler)(request)

    def _get_handler(self, request_type: type[Request[Res]]) -> IHandler[Request[Res], Res]:
        if request_type not in self._handlers:
            raise HandlerNotFoundError(request_type)

        return self._handlers[request_type]



"""Interfaces package."""

from .dispatcher import IDispatcher
from .handlers import ICommandHandler, IEventHandler, IHandler, IQueryHandler
from .middleware import IMiddleware
from .observer import IObserver

__all__ = (
    "IDispatcher",
    "IObserver",
    "IMiddleware",
    "ICommandHandler",
    "IEventHandler",
    "IHandler",
    "IQueryHandler",
)



"""Observers package.

Available:
    - Event
    - Base

Base is useful for creating own observer.
"""

from .base import BaseObserverImpl
from .event import EventObserverImpl

__all__ = ("BaseObserverImpl", "EventObserverImpl")



from dataclasses import dataclass


@dataclass(eq=False)
class BaseError(Exception):
    @property
    def message(self) -> str:
        return "Base error"

    def __str__(self) -> str:
        return self.message



import abc
import typing as tp

from meator.entities import Request
from meator.interfaces.handlers.request import IHandler


class IMiddleware(abc.ABC):
    @abc.abstractmethod
    async def __call__(self, call_next: IHandler, request: Request) -> tp.Any: ...



import abc
import typing as tp

from meator.entities import Request

HRes = tp.TypeVar("HRes")
RType = tp.TypeVar("RType", bound=Request)


class IHandler(abc.ABC, tp.Generic[RType, HRes]):
    @abc.abstractmethod
    async def __call__(self, request: RType) -> HRes: ...



import typing as tp
from typing import TypeVar

from meator._constants import _sentinel
from meator._utils.sentinel_default import get_default_sentinel
from meator.entities.request import Request
from meator.interfaces.handlers.request import IHandler
from meator.interfaces.middleware import IMiddleware
from meator.interfaces.observer import IObserver
from meator.middlewares import wrap_handler

Res = TypeVar("Res")
Req = TypeVar("Req", bound=Request)


class BaseObserverImpl(IObserver):
    def __init__(
        self,
        *,
        middlewares: tp.Sequence[IMiddleware] = _sentinel,
        handlers: dict[type[Request], tp.Sequence[IHandler]] = _sentinel,
    ) -> None:
        self._middlewares = get_default_sentinel(middlewares, ())
        self._handlers = get_default_sentinel(handlers, {})

    def register(self, request: type[Req], handlers: tp.Sequence[IHandler[Req, Res]]) -> None:
        self._handlers[request] = handlers

    async def handle(self, request: Request[None]) -> None:
        handlers = self._get_handler(request)
        for handler in handlers:
            await wrap_handler.wrap_handler_with_middleware(self._middlewares, handler)(request)

    def _get_handler(self, request: Request[Res]) -> tp.Sequence[IHandler[Request, Res]]:
        return self._handlers.get(type(request), ())



from dataclasses import dataclass

from meator.entities import Request
from meator.errors import BaseError


class DispatcherError(BaseError): ...


@dataclass(eq=False)
class HandlerNotFoundError(DispatcherError):
    request_type: type[Request]

    @property
    def message(self) -> str:
        return f"Request handler for {self.request_type.__name__} request is not registered"



import functools
import typing as tp

from meator.entities import Request
from meator.interfaces.handlers.request import IHandler
from meator.interfaces.middleware import IMiddleware

Res = tp.TypeVar("Res")


def wrap_handler_with_middleware(
    middlewares: tp.Sequence[IMiddleware],
    handler: IHandler,
) -> tp.Callable[[Request[Res]], tp.Awaitable[Res]]:
    call_next: tp.Callable[[Request[Res]], tp.Awaitable[Res]] = handler
    for middleware in middlewares:
        call_next = functools.partial(middleware, call_next)

    return call_next



"""Dispatchers package.

Available:
    - Command
    - Query
    - Base

Base is useful for creating own dispatcher.
"""

from .base import BaseDispatcherImpl
from .command import CommandDispatcherImpl
from .query import QueryDispatcherImpl

__all__ = ("BaseDispatcherImpl", "CommandDispatcherImpl", "QueryDispatcherImpl")
